@@ -52,8 +52,11 @@
  * 
  * #GDesktopAppInfo is an implementation of #GAppInfo based on
  * desktop files.
- *
- **/
+ * 
+ * Note that GDesktopAppInfo belongs to the UNIX-specific GIO interfaces,
+ * thus you have to use the <filename>gio-unix-2.0.pc</filename> pkg-config
+ * file when using it.
+ */
 
 #define DEFAULT_APPLICATIONS_GROUP  "Default Applications" 
 #define ADDED_ASSOCIATIONS_GROUP    "Added Associations" 
@@ -267,7 +270,18 @@ g_desktop_app_info_new_from_keyfile (GKeyFile *key_file)
 	  g_object_unref (file);
 	}
       else
-	info->icon = g_themed_icon_new (info->icon_name);
+        {
+          char *p;
+
+          /* Work around a common mistake in desktop files */    
+          if ((p = strrchr (info->icon_name, '.')) != NULL &&
+              (strcmp (p, ".png") == 0 ||
+               strcmp (p, ".xpm") == 0 ||
+               strcmp (p, ".svg") == 0)) 
+            *p = 0;
+
+	  info->icon = g_themed_icon_new (info->icon_name);
+        }
     }
   
   if (info->exec)
@@ -468,6 +482,14 @@ g_desktop_app_info_get_executable (GAppInfo *appinfo)
   return info->binary;
 }
 
+static const char *
+g_desktop_app_info_get_commandline (GAppInfo *appinfo)
+{
+  GDesktopAppInfo *info = G_DESKTOP_APP_INFO (appinfo);
+  
+  return info->exec;
+}
+
 static GIcon *
 g_desktop_app_info_get_icon (GAppInfo *appinfo)
 {
@@ -1284,7 +1306,7 @@ update_mimeapps_list (const char  *desktop_id,
 	     add the current list of (not yet listed) apps before it. */
 
 	  list[i] = NULL; /* Terminate current list so we can use it */
-	  system_list =  get_all_desktop_entries_for_mime_type (content_type, list);
+	  system_list =  get_all_desktop_entries_for_mime_type (content_type, (const char **)list);
 	  
 	  list = g_renew (char *, list, 1 + length + g_list_length (system_list) + 1);
 	  
@@ -1532,8 +1554,8 @@ g_desktop_app_info_remove_supports_type (GAppInfo    *appinfo,
 }
 
 static gboolean
-g_desktop_app_info_ensure_saved (GDesktopAppInfo *info,
-				 GError **error)
+g_desktop_app_info_ensure_saved (GDesktopAppInfo  *info,
+				 GError          **error)
 {
   GKeyFile *key_file;
   char *dirname;
@@ -1628,7 +1650,10 @@ g_desktop_app_info_can_delete (GAppInfo *appinfo)
   GDesktopAppInfo *info = G_DESKTOP_APP_INFO (appinfo);
 
   if (info->filename)
-    return g_access (info->filename, W_OK) == 0;
+    {
+      if (strstr (info->filename, "/userapp-"))
+        return g_access (info->filename, W_OK) == 0;
+    }
 
   return FALSE;
 }
@@ -1731,6 +1756,7 @@ g_desktop_app_info_iface_init (GAppInfoIface *iface)
   iface->remove_supports_type = g_desktop_app_info_remove_supports_type;
   iface->can_delete = g_desktop_app_info_can_delete;
   iface->do_delete = g_desktop_app_info_delete;
+  iface->get_commandline = g_desktop_app_info_get_commandline;
 }
 
 static gboolean
@@ -1790,7 +1816,7 @@ g_app_info_get_all_for_type (const char *content_type)
 
 /**
  * g_app_info_reset_type_associations:
- * content_type: a content type 
+ * @content_type: a content type 
  *
  * Removes all changes to the type associations done by
  * g_app_info_set_as_default_for_type(), 
