@@ -3,6 +3,7 @@
  *
  * g_atomic_*: atomic operations.
  * Copyright (C) 2003 Sebastian Wilhelmi
+ * Copyright (C) 2007 Nokia Corporation
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -19,11 +20,15 @@
  * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
  * Boston, MA 02111-1307, USA.
  */
- 
+
 #include "config.h"
 
+#if defined (G_ATOMIC_ARM)
+#include <sched.h>
+#endif
+
 #include "glib.h"
-#include "gthreadinit.h"
+#include "gthreadprivate.h"
 #include "galias.h"
 
 #if defined (__GNUC__)
@@ -276,6 +281,7 @@ g_atomic_int_exchange_and_add (volatile gint *atomic,
 			       gint           val)
 {
   gint result, temp;
+#if ASM_NUMERIC_LABELS
   __asm__ __volatile__ ("1:       lwarx   %0,0,%3\n"
 			"         add     %1,%0,%4\n"
 			"         stwcx.  %1,0,%3\n"
@@ -283,6 +289,15 @@ g_atomic_int_exchange_and_add (volatile gint *atomic,
 			: "=&b" (result), "=&r" (temp), "=m" (*atomic)
 			: "b" (atomic), "r" (val), "m" (*atomic)
 			: "cr0", "memory");
+#else
+  __asm__ __volatile__ (".Lieaa%=:       lwarx   %0,0,%3\n"
+			"         add     %1,%0,%4\n"
+			"         stwcx.  %1,0,%3\n"
+			"         bne-    .Lieaa%="
+			: "=&b" (result), "=&r" (temp), "=m" (*atomic)
+			: "b" (atomic), "r" (val), "m" (*atomic)
+			: "cr0", "memory");
+#endif
   return result;
 }
  
@@ -292,6 +307,7 @@ g_atomic_int_add (volatile gint *atomic,
 		  gint           val)
 {
   gint result, temp;  
+#if ASM_NUMERIC_LABELS
   __asm__ __volatile__ ("1:       lwarx   %0,0,%3\n"
 			"         add     %1,%0,%4\n"
 			"         stwcx.  %1,0,%3\n"
@@ -299,6 +315,15 @@ g_atomic_int_add (volatile gint *atomic,
 			: "=&b" (result), "=&r" (temp), "=m" (*atomic)
 			: "b" (atomic), "r" (val), "m" (*atomic)
 			: "cr0", "memory");
+#else
+  __asm__ __volatile__ (".Lia%=:       lwarx   %0,0,%3\n"
+			"         add     %1,%0,%4\n"
+			"         stwcx.  %1,0,%3\n"
+			"         bne-    .Lia%="
+			: "=&b" (result), "=&r" (temp), "=m" (*atomic)
+			: "b" (atomic), "r" (val), "m" (*atomic)
+			: "cr0", "memory");
+#endif
 }
 #   else /* !__OPTIMIZE__ */
 gint
@@ -331,6 +356,7 @@ g_atomic_int_compare_and_exchange (volatile gint *atomic,
 				   gint           newval)
 {
   gint result;
+#if ASM_NUMERIC_LABELS
   __asm__ __volatile__ ("sync\n"
 			"1: lwarx   %0,0,%1\n"
 			"   subf.   %0,%2,%0\n"
@@ -341,6 +367,18 @@ g_atomic_int_compare_and_exchange (volatile gint *atomic,
 			: "=&r" (result)
 			: "b" (atomic), "r" (oldval), "r" (newval)
 			: "cr0", "memory"); 
+#else
+  __asm__ __volatile__ ("sync\n"
+			".L1icae%=: lwarx   %0,0,%1\n"
+			"   subf.   %0,%2,%0\n"
+			"   bne     .L2icae%=\n"
+			"   stwcx.  %3,0,%1\n"
+			"   bne-    .L1icae%=\n"
+			".L2icae%=: isync"
+			: "=&r" (result)
+			: "b" (atomic), "r" (oldval), "r" (newval)
+			: "cr0", "memory"); 
+#endif
   return result == 0;
 }
 
@@ -350,6 +388,7 @@ g_atomic_pointer_compare_and_exchange (volatile gpointer *atomic,
 				       gpointer           newval)
 {
   gpointer result;
+#if ASM_NUMERIC_LABELS
   __asm__ __volatile__ ("sync\n"
 			"1: lwarx   %0,0,%1\n"
 			"   subf.   %0,%2,%0\n"
@@ -360,6 +399,18 @@ g_atomic_pointer_compare_and_exchange (volatile gpointer *atomic,
 			: "=&r" (result)
 			: "b" (atomic), "r" (oldval), "r" (newval)
 			: "cr0", "memory"); 
+#else
+  __asm__ __volatile__ ("sync\n"
+			".L1pcae%=: lwarx   %0,0,%1\n"
+			"   subf.   %0,%2,%0\n"
+			"   bne     .L2pcae%=\n"
+			"   stwcx.  %3,0,%1\n"
+			"   bne-    .L1pcae%=\n"
+			".L2pcae%=: isync"
+			: "=&r" (result)
+			: "b" (atomic), "r" (oldval), "r" (newval)
+			: "cr0", "memory"); 
+#endif
   return result == 0;
 }
 #   elif GLIB_SIZEOF_VOID_P == 8 /* 64-bit system */
@@ -369,6 +420,7 @@ g_atomic_int_compare_and_exchange (volatile gint *atomic,
 				   gint           newval)
 {
   gpointer result;
+#if ASM_NUMERIC_LABELS
   __asm__ __volatile__ ("sync\n"
 			"1: lwarx   %0,0,%1\n"
 			"   extsw   %0,%0\n"
@@ -380,6 +432,19 @@ g_atomic_int_compare_and_exchange (volatile gint *atomic,
 			: "=&r" (result)
 			: "b" (atomic), "r" (oldval), "r" (newval)
 			: "cr0", "memory"); 
+#else
+  __asm__ __volatile__ ("sync\n"
+			".L1icae%=: lwarx   %0,0,%1\n"
+			"   extsw   %0,%0\n"
+			"   subf.   %0,%2,%0\n"
+			"   bne     .L2icae%=\n"
+			"   stwcx.  %3,0,%1\n"
+			"   bne-    .L1icae%=\n"
+			".L2icae%=: isync"
+			: "=&r" (result)
+			: "b" (atomic), "r" (oldval), "r" (newval)
+			: "cr0", "memory"); 
+#endif
   return result == 0;
 }
 
@@ -389,6 +454,7 @@ g_atomic_pointer_compare_and_exchange (volatile gpointer *atomic,
 				       gpointer           newval)
 {
   gpointer result;
+#if ASM_NUMERIC_LABELS
   __asm__ __volatile__ ("sync\n"
 			"1: ldarx   %0,0,%1\n"
 			"   subf.   %0,%2,%0\n"
@@ -399,6 +465,18 @@ g_atomic_pointer_compare_and_exchange (volatile gpointer *atomic,
 			: "=&r" (result)
 			: "b" (atomic), "r" (oldval), "r" (newval)
 			: "cr0", "memory"); 
+#else
+  __asm__ __volatile__ ("sync\n"
+			".L1pcae%=: ldarx   %0,0,%1\n"
+			"   subf.   %0,%2,%0\n"
+			"   bne     .L2pcae%=\n"
+			"   stdcx.  %3,0,%1\n"
+			"   bne-    .L1pcae%=\n"
+			".L2pcae%=: isync"
+			: "=&r" (result)
+			: "b" (atomic), "r" (oldval), "r" (newval)
+			: "cr0", "memory"); 
+#endif
   return result == 0;
 }
 #  else /* What's that */
@@ -482,9 +560,222 @@ g_atomic_pointer_compare_and_exchange (volatile gpointer *atomic,
 #  else /* What's that */
 #    error "Your system has an unsupported pointer size"
 #  endif /* GLIB_SIZEOF_VOID_P */
-# else /* !G_ATOMIC_IA64 */
+# elif defined (G_ATOMIC_ARM)
+static volatile int atomic_spin = 0;
+
+static int atomic_spin_trylock (void)
+{
+  int result;
+
+  asm volatile (
+    "swp %0, %1, [%2]\n"
+    : "=&r,&r" (result)
+    : "r,0" (1), "r,r" (&atomic_spin)
+    : "memory");
+  if (result == 0)
+    return 0;
+  else
+    return -1;
+}
+
+static void atomic_spin_lock (void)
+{
+  while (atomic_spin_trylock())
+    sched_yield();
+}
+
+static void atomic_spin_unlock (void)
+{
+  atomic_spin = 0;
+}
+
+gint
+g_atomic_int_exchange_and_add (volatile gint *atomic, 
+			       gint           val)
+{
+  gint result;
+ 
+  atomic_spin_lock();  
+  result = *atomic;
+  *atomic += val;
+  atomic_spin_unlock();
+
+  return result;
+}
+
+void
+g_atomic_int_add (volatile gint *atomic,
+		  gint           val)
+{
+  atomic_spin_lock();
+  *atomic += val;
+  atomic_spin_unlock();
+}
+
+gboolean
+g_atomic_int_compare_and_exchange (volatile gint *atomic, 
+				   gint           oldval, 
+				   gint           newval)
+{
+  gboolean result;
+
+  atomic_spin_lock();
+  if (*atomic == oldval)
+    {
+      result = TRUE;
+      *atomic = newval;
+    }
+  else
+    result = FALSE;
+  atomic_spin_unlock();
+
+  return result;
+}
+
+gboolean
+g_atomic_pointer_compare_and_exchange (volatile gpointer *atomic, 
+				       gpointer           oldval, 
+				       gpointer           newval)
+{
+  gboolean result;
+ 
+  atomic_spin_lock();
+  if (*atomic == oldval)
+    {
+      result = TRUE;
+      *atomic = newval;
+    }
+  else
+    result = FALSE;
+  atomic_spin_unlock();
+
+  return result;
+}
+# elif defined (G_ATOMIC_CRIS) || defined (G_ATOMIC_CRISV32)
+#  ifdef G_ATOMIC_CRIS
+#   define CRIS_ATOMIC_INT_CMP_XCHG(atomic, oldval, newval)		\
+  ({									\
+     gboolean __result;							\
+     __asm__ __volatile__ ("\n"						\
+                           "0:\tclearf\n\t"				\
+                           "cmp.d [%[Atomic]], %[OldVal]\n\t"		\
+                           "bne 1f\n\t"					\
+                           "ax\n\t"					\
+                           "move.d %[NewVal], [%[Atomic]]\n\t"		\
+                           "bwf 0b\n"					\
+                           "1:\tseq %[Result]"				\
+                           : [Result] "=&r" (__result),			\
+                                      "=m" (*(atomic))			\
+                           : [Atomic] "r" (atomic),			\
+                             [OldVal] "r" (oldval),			\
+                             [NewVal] "r" (newval),			\
+                                      "g" (*(gpointer*) (atomic))	\
+                           : "memory");					\
+     __result;								\
+  })
+#  else
+#   define CRIS_ATOMIC_INT_CMP_XCHG(atomic, oldval, newval)		\
+  ({									\
+     gboolean __result;							\
+     __asm__ __volatile__ ("\n"						\
+                           "0:\tclearf p\n\t"				\
+                           "cmp.d [%[Atomic]], %[OldVal]\n\t"		\
+                           "bne 1f\n\t"					\
+                           "ax\n\t"					\
+                           "move.d %[NewVal], [%[Atomic]]\n\t"		\
+                           "bcs 0b\n"					\
+                           "1:\tseq %[Result]"				\
+                           : [Result] "=&r" (__result),			\
+                                      "=m" (*(atomic))			\
+                           : [Atomic] "r" (atomic),			\
+                             [OldVal] "r" (oldval),			\
+                             [NewVal] "r" (newval),			\
+                                      "g" (*(gpointer*) (atomic))	\
+                           : "memory");					\
+     __result;								\
+  })
+#  endif
+
+#define CRIS_CACHELINE_SIZE 32
+#define CRIS_ATOMIC_BREAKS_CACHELINE(atomic) \
+  (((gulong)(atomic) & (CRIS_CACHELINE_SIZE - 1)) > (CRIS_CACHELINE_SIZE - sizeof (atomic)))
+
+gint     __g_atomic_int_exchange_and_add         (volatile gint   *atomic,
+						  gint             val);
+void     __g_atomic_int_add                      (volatile gint   *atomic,
+						  gint             val);
+gboolean __g_atomic_int_compare_and_exchange     (volatile gint   *atomic,
+						  gint             oldval,
+						  gint             newval);
+gboolean __g_atomic_pointer_compare_and_exchange (volatile gpointer *atomic,
+						  gpointer         oldval,
+						  gpointer         newval);
+
+gboolean
+g_atomic_pointer_compare_and_exchange (volatile gpointer *atomic,
+				       gpointer           oldval,
+				       gpointer           newval)
+{
+  if (G_UNLIKELY (CRIS_ATOMIC_BREAKS_CACHELINE (atomic)))
+    return __g_atomic_pointer_compare_and_exchange (atomic, oldval, newval);
+
+  return CRIS_ATOMIC_INT_CMP_XCHG (atomic, oldval, newval);
+}
+
+gboolean
+g_atomic_int_compare_and_exchange (volatile gint *atomic,
+				   gint           oldval,
+				   gint           newval)
+{
+  if (G_UNLIKELY (CRIS_ATOMIC_BREAKS_CACHELINE (atomic)))
+    return __g_atomic_int_compare_and_exchange (atomic, oldval, newval);
+
+  return CRIS_ATOMIC_INT_CMP_XCHG (atomic, oldval, newval);
+}
+
+gint
+g_atomic_int_exchange_and_add (volatile gint *atomic,
+			       gint           val)
+{
+  gint result;
+
+  if (G_UNLIKELY (CRIS_ATOMIC_BREAKS_CACHELINE (atomic)))
+    return __g_atomic_int_exchange_and_add (atomic, val);
+
+  do
+    result = *atomic;
+  while (!CRIS_ATOMIC_INT_CMP_XCHG (atomic, result, result + val));
+
+  return result;
+}
+
+void
+g_atomic_int_add (volatile gint *atomic,
+		  gint           val)
+{
+  gint result;
+
+  if (G_UNLIKELY (CRIS_ATOMIC_BREAKS_CACHELINE (atomic)))
+    return __g_atomic_int_add (atomic, val);
+
+  do
+    result = *atomic;
+  while (!CRIS_ATOMIC_INT_CMP_XCHG (atomic, result, result + val));
+}
+
+/* We need the atomic mutex for atomic operations where the atomic variable
+ * breaks the 32 byte cache line since the CRIS architecture does not support
+ * atomic operations on such variables. Fortunately this should be rare.
+ */
 #  define DEFINE_WITH_MUTEXES
-# endif /* G_ATOMIC_IA64 */
+#  define g_atomic_int_exchange_and_add __g_atomic_int_exchange_and_add
+#  define g_atomic_int_add __g_atomic_int_add
+#  define g_atomic_int_compare_and_exchange __g_atomic_int_compare_and_exchange
+#  define g_atomic_pointer_compare_and_exchange __g_atomic_pointer_compare_and_exchange
+
+# else /* !G_ATOMIC_* */
+#  define DEFINE_WITH_MUTEXES
+# endif /* G_ATOMIC_* */
 #else /* !__GNUC__ */
 # ifdef G_PLATFORM_WIN32
 #  define DEFINE_WITH_WIN32_INTERLOCKED
